// opttarget.cpp : Defines the entry point for the console application.
//
//#define WINVER 0x0501
#include "stdafx.h"
#include <Windows.h>
#include <xmmintrin.h>
#include <emmintrin.h>
#include "opttarget.h"

#define ERR_CANTOPEN -1
#define ERR_BADFILE  -2

int ReadBMP(const char* fname, int &X, int &Y, int &bpp, BYTE *pData, BITMAPINFOHEADER **ppbmi = NULL, DWORD* pbmisz = NULL)
{
	int ret = 0;
	DWORD bmisize = 0;
	BITMAPINFOHEADER *pbmi = NULL;
	FILE *f = fopen(fname,"rb");
	if (!f)
		return ERR_CANTOPEN;
	BITMAPFILEHEADER hdr;
	memset(&hdr, 0, sizeof(BITMAPFILEHEADER));
	size_t len = fread(&hdr, 1, sizeof(BITMAPFILEHEADER), f);
	if (len != sizeof(BITMAPFILEHEADER) || hdr.bfType != 0x4d42) {
		ret = ERR_BADFILE;
		goto end;
	}
	if (hdr.bfSize < sizeof(BITMAPFILEHEADER)) 
		fseek(f, hdr.bfSize, SEEK_SET);

	bmisize = hdr.bfOffBits - sizeof(BITMAPFILEHEADER);
	pbmi = (BITMAPINFOHEADER*)calloc(1, bmisize);
	len = fread(pbmi, 1, bmisize, f); //read the rest
	if (len != bmisize) {
		ret = ERR_BADFILE;
		goto end;
	}

	if (ppbmi) {
		*ppbmi = (BITMAPINFOHEADER*)malloc(bmisize);
		memcpy(*ppbmi, pbmi, bmisize);
		*pbmisz = bmisize;
	}

	X = pbmi->biWidth; Y = abs(pbmi->biHeight); bpp = pbmi->biBitCount;
	if (pData) {
		len = fread(pData, bpp/8, X*Y, f);
		if (len != X*Y) {
			printf("error reading the bits!\n");
			ret = ERR_BADFILE;
		}
	}

	free(pbmi);
end:
	fclose(f);
	return ret;
}


template<class T>
void ClipT(RGB<T> &x)
{
	x.clip();
}

void ClipT(int &x)
{
	if (x<0)
		x = 0;
	if (x>255)
		x = 255;
}


static int QPK[4][4] = {
	{0, 128, 0, 0},
	{-7, 105, 35, -5},
	{-8, 72, 72, -8},
	{-5, 35, 105, -7} };
//x,y in whole pixels, dx,dy in quarters 
template<class T, int border, class GT, int W>
void GetBlock4(T * const pDat, const int X, const int Y, 
			   T* bldt, int x, int y, int dx, int dy) 
{
	const int w = X;
	const int blw = W;
	x += dx>>2; y += dy>>2;
	dx &= 3; dy &= 3;
	const int * const qkx=QPK[dx];
	const int * const qky=QPK[dy];
	const int off = border*w + border;

	if (!(dx+dy)) {
		//GetBlock(bl, x,y);
		int i = y*w + x + off, j=0;		
		for(int yy=0;yy<blw*blw;yy+=blw, i+=w) 
			memcpy(&bldt[yy], &pDat[i], blw * sizeof(T) ); 
		return;
	}
	
	int k=0;
	if (x < -border+1) x = -border+1;
	if (y < -border+1) y = -border+1;
	if ((x + blw) > (X + border-2)) x = X + border-2 - blw;
	if ((y + blw) > (Y + border-2)) y = Y + border-2 - blw;
	//T *bldt = bl->pData;
	//T * const pDat = pData;
	

	if (dx && !dy) {
		int si = (y+0)*w + x + off -1, dsi = w - blw;	
		#pragma prefetch pDat
		for(int i=0;i<blw;i++) {
			//_mm_prefetch((char*)&pDat[si+w], _MM_HINT_T0);
			for(int j=0;j<blw;j++) {
				GT t = ( (GT)pDat[si]*qkx[0] + (GT)pDat[si+1]*qkx[1] + (GT)pDat[si+2]*qkx[2] + (GT)pDat[si+3]*qkx[3] +64)>>7;
				ClipT(t);
				bldt[k+j] = t;
				si++;
			}
			k += blw;
			si += dsi;
		}
		return;
	}
	if (!dx && dy) {
		int si = (y+0)*w + x + off, dsi = w - blw;
		#pragma prefetch pDat
		for(int i=0;i<blw;i++) {
			for(int j=0;j<blw;j++) {
				GT t = ( (GT)pDat[si-w]*qky[0] + (GT)pDat[si]*qky[1] + (GT)pDat[si+w]*qky[2] + (GT)pDat[si+2*w]*qky[3] +64)>>7;
				ClipT(t);
				bldt[k+j] = t;
				si++;
			}
			si += dsi;
			k += blw;
		}
		return;
	}

	//here dx > 0 and dy > 0		
	__m128i q32, q10;
	switch(dx) {
	case 0:
		q10 = _mm_set_epi16(128, 0, 128, 0, 128, 0, 128, 0);
		q32 = _mm_set_epi16(0,0, 0,0, 0,0, 0,0);
		break;
	case 1:
		q10 = _mm_set_epi16(105, -7, 105, -7, 105, -7, 105, -7);
		q32 = _mm_set_epi16(-5, 35, -5, 35, -5, 35, -5, 35);
		break;
	case 2:
		q10 = _mm_set_epi16(72, -8, 72, -8, 72, -8, 72, -8);
		q32 = _mm_set_epi16(-8, 72, -8, 72, -8, 72, -8, 72);
		break;
	case 3:
		q10 = _mm_set_epi16(35, -5, 35, -5, 35, -5, 35, -5);
		q32 = _mm_set_epi16(-7, 105, -7, 105, -7, 105, -7, 105);
		break;
	}
	const __m128i zero = _mm_setzero_si128();
	const __m128i one = _mm_set1_epi32(1);
	__declspec(align(16)) GT gbdata[W * (W+4) + 16]; //16*20+16
	GT *pDt = gbdata; const int gboff = W;
	int si = (y-1)*w + x + off - 1, dsi = w - blw;

	
	//_mm_prefetch((char*)&pDat[si-1], _MM_HINT_T0);
	k = 0;
	#pragma prefetch pDat
	for(int i=-1;i<blw+2;i++) {	//i in [-1..17]	19 iters
		//_mm_prefetch((char*)&pDat[si+w], _MM_HINT_T0);
		const int kend = k + W; //k+16
		for(;k<kend;k++) {
			//pDt[k] = ( (GT)pDat[si]*qkx[0] + (GT)pDat[si+1]*qkx[1] + (GT)pDat[si+2]*qkx[2] + (GT)pDat[si+3]*qkx[3] +1)>>1;
			__m128i rgb4rgb3rgb2rgb1rgb0_u8 = _mm_loadu_si128((const __m128i*)&pDat[si]);
			__m128i rgb4rgb3rgb2_u8 = _mm_srli_si128(rgb4rgb3rgb2rgb1rgb0_u8, 6);
			__m128i rgb1rgb0_i16 = _mm_unpacklo_epi8(rgb4rgb3rgb2rgb1rgb0_u8, zero);
			__m128i rgb3rgb2_i16 = _mm_unpacklo_epi8(rgb4rgb3rgb2_u8, zero);
			__m128i rgb1_i16 = _mm_srli_si128(rgb1rgb0_i16, 6);
			__m128i rgb3_i16 = _mm_srli_si128(rgb3rgb2_i16, 6);
			__m128i r1r0g1g0b1b0_i16 = _mm_unpacklo_epi16(rgb1rgb0_i16, rgb1_i16);
			__m128i r3r2g3g2b3b2_i16 = _mm_unpacklo_epi16(rgb3rgb2_i16, rgb3_i16);
			__m128i r10g10b10_i32 = _mm_madd_epi16(r1r0g1g0b1b0_i16, q10);
			__m128i r32g32b32_i32 = _mm_madd_epi16(r3r2g3g2b3b2_i16, q32);
			__m128i r3210g3210b3210_i32 = _mm_add_epi32(r10g10b10_i32, r32g32b32_i32);
			__m128i bplus1_i32 = _mm_add_epi32(r3210g3210b3210_i32, one);
			__m128i res_i32 = _mm_srai_epi32(bplus1_i32, 1);
			_mm_storeu_si128((__m128i*)&pDt[k], res_i32);
			si++;
		}
		si += dsi;
		//k += blw;
	} // last pDt write at [18*16+15]

	const __m128i half = _mm_set1_epi32(4096);
	const __m128i c255 = _mm_set_epi16(0,255, 0,255, 0,255, 0,255);
	__declspec(align(16)) int ires[4];

	switch(dy) {
	case 0:
		q10 = _mm_set_epi16(128, 0, 128, 0, 128, 0, 128, 0);
		q32 = _mm_set_epi16(0,0, 0,0, 0,0, 0,0);
		break;
	case 1:
		q10 = _mm_set_epi16(105, -7, 105, -7, 105, -7, 105, -7);
		q32 = _mm_set_epi16(-5, 35, -5, 35, -5, 35, -5, 35);
		break;
	case 2:
		q10 = _mm_set_epi16(72, -8, 72, -8, 72, -8, 72, -8);
		q32 = _mm_set_epi16(-8, 72, -8, 72, -8, 72, -8, 72);
		break;
	case 3:
		q10 = _mm_set_epi16(35, -5, 35, -5, 35, -5, 35, -5);
		q32 = _mm_set_epi16(-7, 105, -7, 105, -7, 105, -7, 105);
		break;
	}

	k = 0;
	si = 0;	//dsi = w2 - blw;
	
	#pragma prefetch pDt
	for(int i=0;i<blw;i++) { // i in [0..15]
		const int kend = k + W; // + 16
		for(;k<kend;k++) {
			//GT t = (pDt[si]*qky[0] + pDt[si+W]*qky[1] + pDt[si+W*2]*qky[2] +pDt[si+W*3]*qky[3] +4096)>>13;//+64)>>7;
			//ClipT(t);
			//bldt[k] = t;
			__m128i rgb0_i32 = _mm_loadu_si128((const __m128i*)&pDt[si]);
			__m128i rgb1_i32 = _mm_loadu_si128((const __m128i*)&pDt[si+W]);
			__m128i rgb0_i16 = _mm_packs_epi32(rgb0_i32, zero);
			__m128i rgb1_i16 = _mm_packs_epi32(rgb1_i32, zero);
			__m128i r1r0g1g0b1b0_i16 = _mm_unpacklo_epi16(rgb0_i16, rgb1_i16);
			__m128i rgb2_i32 = _mm_loadu_si128((const __m128i*)&pDt[si+W*2]);
			__m128i rgb3_i32 = _mm_loadu_si128((const __m128i*)&pDt[si+W*3]);
			__m128i rgb2_i16 = _mm_packs_epi32(rgb2_i32, zero);
			__m128i rgb3_i16 = _mm_packs_epi32(rgb3_i32, zero);
			__m128i r3r2g3g2b3b2_i16 = _mm_unpacklo_epi16(rgb2_i16, rgb3_i16);
			__m128i r10g10b10_i32 = _mm_madd_epi16(r1r0g1g0b1b0_i16, q10);
			__m128i r32g32b32_i32 = _mm_madd_epi16(r3r2g3g2b3b2_i16, q32);
			__m128i r3210g3210b3210_i32 = _mm_add_epi32(r10g10b10_i32, r32g32b32_i32);
			__m128i bplus1_i32 = _mm_add_epi32(r3210g3210b3210_i32, half);
			__m128i sum_shifted = _mm_srai_epi32(bplus1_i32, 13);
			__m128i res0 = _mm_max_epi16(sum_shifted, zero);
			__m128i res_i32 = _mm_min_epi16(res0, c255);
			__m128i res_i16 = _mm_packs_epi32(res_i32, zero);
			__m128i res_i8 = _mm_packus_epi16(res_i16, zero);
			_mm_storel_epi64((__m128i*)ires, res_i8);			
			//bldt[k] = RGB<BYTE>(ires[2], ires[1], ires[0]);
			memcpy(&bldt[k], ires, 3);
			si++;
		}
		//si += dsi;
		//k += blw;
	}	
}


int _tmain(int argc, _TCHAR* argv[])
{
	int X,Y, bpp; 
    BITMAPINFOHEADER *pbmi = NULL;
	DWORD bmisz = 0;
	const char *filename = "picture.bmp";
	int res = ReadBMP(filename, X,Y,bpp, NULL, &pbmi, &bmisz);
	if (res<0) {
		printf("error %d!\n", res);
		return 1;
	}
	const int sz = X * Y * bpp / 8;
	BYTE *pData = new BYTE[sz];
	ReadBMP(filename, X,Y,bpp, pData); 

	const int W = 16;
	RGB<BYTE> block[W*W];	

	const int bsz = W*W*3;
	char dstname[512];
	
	BITMAPFILEHEADER hdr;
	memset(&hdr, 0, sizeof(BITMAPFILEHEADER));
	hdr.bfType = 0x4d42;
	hdr.bfSize = sizeof(BITMAPFILEHEADER) + bmisz + bsz;
	hdr.bfOffBits = sizeof(BITMAPFILEHEADER) + bmisz;
	pbmi->biHeight = W;
	pbmi->biWidth = W;
	pbmi->biSizeImage = W*W*3;

	for(int n=100;n<2000;n++) {
		sprintf(dstname, "block%.4d.bmp", n);
		FILE *f = fopen(dstname, "wb");
		if (f) {
			GetBlock4<RGB<BYTE>, 16, RGB<int>, W>((RGB<BYTE>*)pData, X, Y, block, n/4,n/4, n % 4, n % 4);
			fwrite(&hdr, 1, sizeof(BITMAPFILEHEADER), f);
			fwrite(pbmi, 1, bmisz, f);
			fwrite(block, 1, bsz, f);
			fclose(f);
		}
	}

	return 0;
}

