#define SCLIP(x) (x < 0 ? 0 : x > 255 ? 255 : x)

template<class T>
class RGB {
public:
	T b,g,r; ///<member values

	///Empty constructor.
	RGB<T>() {};
	///Explicit constructor.
	RGB<T>(T r_, T g_, T b_) : r(r_), g(g_), b(b_) {}

	RGB<T>(T a) : r(a), g(a), b(a) {}

	///Copy constructor.
	template<class T2>
	RGB<T>( RGB<T2> &a) : r(a.r), g(a.g), b(a.b) {}

	///Load values from BYTE array.
	///@param rgbval points to 3 bytes containing r,g,b values.
	void load(const BYTE *rgbval) {
		b = rgbval[0];		g = rgbval[1];		r = rgbval[2];
	}

	///Copy operator. 
	template<class T2>
	RGB<T>& operator=(RGB<T2> &a) {
		r = a.r;		g = a.g;		b = a.b;
		return *this;
	}

	///Operator +=. Adds values of the parameter to member values.
	template<class T2>
	RGB<T>& operator+=(RGB<T2> &a) {
		r += a.r;		g += a.g;		b += a.b;
		return *this;
	}

	///Operator -=. Subtracts values of the parameter to member values.
	template<class T2>
	RGB<T>& operator-=(RGB<T2> &a) {
		r -= a.r;		g -= a.g;		b -= a.b;
		return *this;
	}

	///Addition. Creates a new RGB value as sum of parameter and member values.
	template<class T2>
	RGB<T> operator+(RGB<T2> &a) const {
		return RGB<T>(r + a.r, g + a.g, b + a.b);
	}
	
	template<class T2>
	RGB<T> operator+(const RGB<T2> &a) const {
		return RGB<T>(r + a.r, g + a.g, b + a.b);
	}

	///Addition of scalar. Adds scalar parameter to all of r,g,b values and returns as new RGB.
	RGB<T> operator+(int a) const {
		return RGB<T>(r + a, g + a, b + a);
	}
	
	///Subtraction. Creates a new RGB value as subtraction of parameter from member values.
	template<class T2>
	RGB<T> operator-(RGB<T2> &a) const {
		return RGB<T>(r - a.r, g - a.g, b - a.b);
	}

	///Multiplication by scalar. Multiplies all member values by integer parameter.
	RGB<T>& operator*=(int a) {
		r *= a;		g *= a;		b *= a;
		return *this;
	}

	///Division by scalar. Divides all member values by integer parameter.
	RGB<T>& operator/=(int a) {
		r /= a;		g /= a;		b /= a;
		return *this;
	}
	
	///Multiplication by scalar. Creates new RGB value multiplicating member values by parameter.
	template<class T2>
	RGB<T> operator*(T2 a) const {
		return RGB<T>(r * a, g * a, b * a);
	}

	///Multiplication by RGB. Creates new RGB value multiplicating member and parameter values one by one.
	template<class T2>
	RGB<T> operator*(RGB<T2> a) {
		return RGB<T>(r * a.r, g * a.g, b * a.b);
	}

	///Division by scalar. Divides all members by integer scalar.
	RGB<T> operator/(int a) const {
		return RGB<T>(r / a, g / a, b / a);
	}
	
	RGB<T> operator>>(int a) const {
		return RGB<T>(r >> a, g >> a, b >> a);
	}
	RGB<T> operator<<(int a) const {
		return RGB<T>(r << a, g << a, b << a);
	}

	///Clipping. Clips member values to be in [0..255].
	void clip() 
	{
		r = ClipT(r);
		g = ClipT(g);
		b = ClipT(b);
	}

	///Calculate Y. Calculates fast Y (luminance) approximation from r,g,b members.
	T CalcY() {
		return ((r+b)/2 + g + 1)/2;
	}

protected:
	///Clipping of scalar. Clips its parameter to be in [0..255]
	T ClipT(T x);

};

template<class T>
T RGB<T>::ClipT(T x)
{
	/*if (x<0)
		return 0;
	if (x>255)
		return 255;
	return x;*/
	return SCLIP(x);
}
